#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <string.h>
#include <sys/file.h>
#include <sys/time.h>
#include "otto_measure.h"

static void usage ()
{
	fprintf (stderr,
		"usage: atmtimes [-I interface] [-s samples] start|stop|read|clear [lovc [hivc]]\n");
	exit (2);
}

main (int argc, char *argv[])
{
	int argn;
	char interface[1024];
	struct otto_measure m;
	int fd;
	struct otto_event *events;
	unsigned vclo = 0;
	unsigned vchi = 0xffffffff;
	int rc = 0;

	strcpy (interface, "/dev/otto0");
	m.vc = 0;
	m.uu.times.maxevents = 1000;

	for (argn = 1; argn < argc && argv[argn][0] == '-'; argn++) {
		char *flags;
		for (flags = &argv[argn][1]; *flags != '\0'; flags++) {
			switch (*flags) {
			case 's':
				if (++argn < argc) {
					m.uu.times.maxevents =
						(unsigned)atoi(argv[argn]);
				} else {
					usage ();
				}
				break;
			case 'I':
				if (++argn < argc) {
					strcpy (interface, "/dev/");
					strcat (interface, argv[argn]);
				} else {
					usage ();
				}
				break;
			default: 
				usage ();
			}
		}
	}

	if (argn >= argc) {
		usage ();
	}

	if (strcmp (argv[argn], "start") == 0) {
		m.cmd = OTTO_T_MK (OTTO_T_START);
	} else if (strcmp (argv[argn], "stop") == 0) {
		m.cmd = OTTO_T_MK (OTTO_T_STOP);
	} else if (strcmp (argv[argn], "read") == 0) {
		m.cmd = OTTO_T_MK (OTTO_T_GET);
		m.uu.times.events = (struct otto_event *)
			malloc (sizeof (struct otto_event) *
			        OTTO_EVENTS_MAX);
		m.uu.times.maxevents = OTTO_EVENTS_MAX;
	} else if (strcmp (argv[argn], "clear") == 0) {
		m.cmd = OTTO_T_MK (OTTO_T_CLEAR);
	} else {
		usage ();
	}
	argn++;

	if (argn < argc) {
		vclo = atoi (argv[argn]);
		argn++;
		if (argn < argc) {
			vchi = atoi (argv[argn]);
			argn++;
		} else {
			vchi = vclo+1;
		}
	}
	m.uu.times.vclo = vclo;
	m.uu.times.vchi = vchi;

	if ((fd = open (interface, O_RDONLY, 0)) < 0) {
		perror (interface);
		exit (1);
	}

	if (ioctl (fd, OTTO_MEASURE, (char *) &m) == -1) {
		perror ("OTTO_MEASURE");
		exit (2);
	}

	if (m.cmd == OTTO_T_MK (OTTO_T_GET)) {
		unsigned i;
		double curcycles;
		double totaltime;
		double totalcycles;
		double conv;
		time_t t;
		char *ct;
		unsigned wraps = 0;
		unsigned lastcycles = m.uu.times.startcycles;

		for (i = 0; i != m.uu.times.nevents; i++) {
			unsigned cycles = m.uu.times.events[i].cycles;
			if (cycles < lastcycles) {
				wraps++;
			}
			lastcycles = cycles;
		}
		if (m.uu.times.endcycles < lastcycles) {
			wraps++;
		}

		totaltime = ((double) m.uu.times.endtime.tv_sec) -
			    ((double) m.uu.times.starttime.tv_sec) +
			    ((double) m.uu.times.endtime.tv_usec -
			     (double) m.uu.times.starttime.tv_usec) / 1.0e6;
		totalcycles = (65536.0 * 65536.0 * (double) wraps) +
				((double) m.uu.times.endcycles) - 
				((double) m.uu.times.startcycles);
		conv = totaltime / (totalcycles+1.0);

		curcycles = 0.0;
		lastcycles = m.uu.times.startcycles;

		t = m.uu.times.starttime.tv_sec;
		ct = ctime (&t);
		printf ("%-10s %12.6f %.8s.%06d\n",
			"start", conv*curcycles,
			&ct[11], m.uu.times.starttime.tv_usec);

		for (i = 0; i != m.uu.times.nevents; i++) {
			struct otto_event *e = &m.uu.times.events[i];
			unsigned et = OTTO_EVENT_TYPE (e->vctype);
			unsigned vc = OTTO_EVENT_VC (e->vctype);
			static char *evtstr[] = {
				"rcv",    "xmt",    "tick",   "event3",
				"event4", "event5", "event6", "event7",
				"event8", "event9", "event10","event11",
				"event12","event13","event14","event15"
			};
			unsigned cycles = m.uu.times.events[i].cycles;

			if (cycles < lastcycles) {
				curcycles += 65536.0 * 65536.0 -
					(double) (lastcycles - cycles);
			} else {
				curcycles += (double) (cycles - lastcycles);
			}

			lastcycles = cycles;

			if (et == OTTO_EVENT_RX || et == OTTO_EVENT_TX) {
				if (vclo <= vc && vc < vchi) {
					printf ("%-10s %12.6f vc %5u len %5u\n",
						evtstr[et], conv*curcycles,
						vc, e->len);
				}
			}
		}

		t = m.uu.times.endtime.tv_sec;
		ct = ctime (&t);
		printf ("%-10s %12.6f %.8s.%06d %s\n",
			"end", conv*curcycles,
			&ct[11], m.uu.times.endtime.tv_usec,
			(m.uu.times.state == OTTO_TIMES_COLLECT)?
			"(still collecting)": "");
		if (m.uu.times.state != OTTO_TIMES_COLLECT) {
			rc = 1;
		}
	}

	(void) close (fd);

	exit (rc);
}
