#include "otto_platform.h"

static struct mbuf *otto_en_atmf_add (struct otto_softc *sc,
	struct mbuf *m, struct sockaddr *dst,
	unsigned char *osrc, unsigned char *odst, int type,
	unsigned vc, int *plen);
static struct mbuf *otto_en_sle_add (struct otto_softc *sc,
	struct mbuf *m, struct sockaddr *dst,
	unsigned char *osrc, unsigned char *odst, int type,
	unsigned vc, int *plen);
static int otto_en_atmf_read (struct otto_softc *sc, struct mbuf *m0,
        unsigned bytes, unsigned len, unsigned vc, int link);

struct otto_en otto_en_atmf = {
	otto_en_atmf_add,
	otto_en_atmf_read
};
struct otto_en otto_en_sle = {
	otto_en_sle_add,
	otto_en_atmf_read
};


static char llchdr[6] = { LLC_SNAP_LSAP,LLC_SNAP_LSAP,LLC_UI,0,0,0 };

#define HDRSIZEATMF (2 + sizeof (struct ether_header) + sizeof (llchdr) + 2)
	/* size of header for atm forum llc encapsulation
	   first 2 bytes = bytes of zeroes in front of Ethernet header
	   second 2 bytes = IEEE copy of Ethernet type field */
#define HDRSIZESLE (2 + sizeof (struct ether_header))
	/* size of header for synoptics encapsulation (same minus llc)
	   first 2 bytes = bytes of zeroes in front of Ethernet header */

static struct mbuf *otto_en_atmf_add (struct otto_softc *sc,
	struct mbuf *m, struct sockaddr *dst,
	unsigned char *osrc, unsigned char *odst, int type,
	unsigned vc, int *plen)
{
	unsigned char *padptr;
	struct ether_header *eh;
	char *lh;
	int len;

	len = m_length (m);
#ifdef __osf__
	M_PREPEND (m, HDRSIZEATMF, M_DONTWAIT);
	if (m == 0) {
		return ((struct mbuf *) 0);
	}
#else
	if (m->m_off > MMAXOFF ||
	    MMINOFF + HDRSIZEATMF > m->m_off ||
	    (m->m_off&0x3)!=0) {
	    	struct mbuf *mtmp;
		OTTOMGET (mtmp);
		if (mtmp == 0) {
			OTTOMFREEM (m);
			return ((struct mbuf *) 0);
		}
		mtmp->m_next = m;
		mtmp->m_off = MMAXOFF;
		mtmp->m_len = 0;
		m = mtmp;
	}
	m->m_off -= HDRSIZEATMF;
	m->m_len += HDRSIZEATMF;
#endif
	padptr = mtod (m, unsigned char *);
	eh = (struct ether_header *) (padptr + 2);
	lh = (char *) &eh[1];
	type = htons ((u_short)type);
        /* build the ethernet header */
        if (((padptr - (unsigned char *) 0) & 1) != 0) {
                /* unaligned case  - let bcopy sort it out */
                padptr[0] = 0;
                padptr[1] = 0;
                bcopy ((char *) odst, (char *) eh->ether_dhost, 6);
                bcopy ((char *) osrc, (char *) eh->ether_shost, 6);
                bcopy (llchdr, lh, sizeof (llchdr));
                bcopy ((char *) &type, &lh[sizeof (llchdr)], 2);

                type = len + sizeof (llchdr) + 2;
                if (type > 1500) {
                        type = 1500;
                }
                type = htons ((u_short)type);
                bcopy ((char *) &type, (char *) &eh->ether_type, 2);
        } else {
                /* aligned case - in-line 2-byte moves */
                *(u_short *)padptr = 0;
                ((u_short *)eh->ether_dhost)[0] = ((u_short *) odst)[0];
                ((u_short *)eh->ether_dhost)[1] = ((u_short *) odst)[1];
                ((u_short *)eh->ether_dhost)[2] = ((u_short *) odst)[2];

                ((u_short *)eh->ether_shost)[0] = ((u_short *) osrc)[0];
                ((u_short *)eh->ether_shost)[1] = ((u_short *) osrc)[1];
                ((u_short *)eh->ether_shost)[2] = ((u_short *) osrc)[2];

                ((u_short *)lh)[0] = ((u_short *) llchdr)[0];
                ((u_short *)lh)[1] = ((u_short *) llchdr)[1];
                ((u_short *)lh)[2] = ((u_short *) llchdr)[2];
                ((u_short *)lh)[3] = type;

                type = len + sizeof (llchdr) + 2;
                if (type > 1500) {
                        type = 1500;
                }
                eh->ether_type = htons ((u_short)type);
        }

	*plen = len + HDRSIZEATMF;
	return (m);
}

static struct mbuf *otto_en_sle_add (struct otto_softc *sc,
	struct mbuf *m, struct sockaddr *dst,
	unsigned char *osrc, unsigned char *odst, int type,
	unsigned vc, int *plen)
{
	unsigned char *padptr;
	struct ether_header *eh;
	int len;

	len = m_length (m);
#ifdef __osf__
	M_PREPEND (m, HDRSIZESLE, M_DONTWAIT);
	if (m == 0) {
		return ((struct mbuf *) 0);
	}
#else
	if (m->m_off > MMAXOFF ||
	    MMINOFF + HDRSIZESLE > m->m_off ||
	    (m->m_off&0x3)!=0) {
	    	struct mbuf *mtmp;
		OTTOMGET (mtmp);
		if (mtmp == 0) {
			OTTOMFREEM (m);
			return ((struct mbuf *) 0);
		}
		mtmp->m_next = m;
		mtmp->m_off = MMAXOFF;
		mtmp->m_len = 0;
		m = mtmp;
	}
	m->m_off -= HDRSIZESLE;
	m->m_len += HDRSIZESLE;
#endif
	padptr = mtod (m, unsigned char *);
	eh = (struct ether_header *) (padptr + 2);
	type = htons ((u_short)type);
        /* build the ethernet header */
        if (((padptr - (unsigned char *) 0) & 1) != 0) {
                /* unaligned case  - let bcopy sort it out */
                padptr[0] = 0;
                padptr[1] = 0;
                bcopy ((char *) odst, (char *) eh->ether_dhost, 6);
                bcopy ((char *) osrc, (char *) eh->ether_shost, 6);
                bcopy ((char *) &type, (char *) &eh->ether_type, 2);
        } else {
                /* aligned case - in-line 2-byte moves */
                *(u_short *)padptr = 0;
                ((u_short *)eh->ether_dhost)[0] = ((u_short *) odst)[0];
                ((u_short *)eh->ether_dhost)[1] = ((u_short *) odst)[1];
                ((u_short *)eh->ether_dhost)[2] = ((u_short *) odst)[2];

                ((u_short *)eh->ether_shost)[0] = ((u_short *) osrc)[0];
                ((u_short *)eh->ether_shost)[1] = ((u_short *) osrc)[1];
                ((u_short *)eh->ether_shost)[2] = ((u_short *) osrc)[2];
                eh->ether_type = type;
        }

	*plen = len + HDRSIZESLE;
	return (m);
}

int otto_nolearn = 0;

static int otto_en_atmf_read (struct otto_softc *sc, struct mbuf *m0,
        unsigned bytes, unsigned len, unsigned vc, int link)
{
        u_char *p = mtod (m0, unsigned char *);
        struct ether_header *eptr = (struct ether_header *) &p[2];
        struct ether_header eh;
        u_short etype;
        int recvthis;
        int hdrsize = 2 + sizeof (struct ether_header);
        int copied = 0;
        unsigned flags;

        if (len < hdrsize) {
        	cprintf ("otto_en_atmf_read: rcv pkt too short len=%d bytes=%d\n",
			len, bytes);
        	OTTO_STATS_INC_L (&sc->data, rcv_pkt_too_short, link);
        	return (0);
        }

        /* check the first two bytes.  They should be zero if this is a
           ROW LAN emulation packet.  If not, give up.  */
        if (p[0] != 0 || p[1] != 0) {
                OTTO_STATS_INC_L (&sc->data, rcv_pkt_not_ethernet, link);
                return (0);
        }

        /* get Ethernet type, network byte order */
        p = (unsigned char *) &eptr->ether_type;
        etype = (p[0] << 8) + p[1];
        if (etype <= ETHERMTU) {
        	p = (unsigned char *) &eptr[1];
                if (bcmp (llchdr, p, sizeof (llchdr)) == 0) {
			/* pick up ethertype */
			p += sizeof (llchdr);
			etype = (p[0] << 8) + p[1];
			hdrsize += sizeof (llchdr) + 2;
		}
        }

        /* recvthis is true if we should receive this packet */
        recvthis = 1;
        if ((eptr->ether_dhost[0] & 1) != 0) {
                /* Multicast filter.
                   This is quoted from the RFC on multicast (1112) section
                   7.3 page 9:
                   >> The local network module must not deliver up any 
                   >> multicast packets that were transmitted from that
                   >> module; loopback of multicasts is handled at the IP
                   >> layer or higher.
                   This occurs even if IFF_PROMISC|IFF_ALLMULTI
                   This is the opposite requirement from broadcast loopbacks.
                 */
                if (bcmp ((caddr_t) eptr->ether_dhost,
                          (caddr_t) etherbroadcastaddr, 6) != 0) {
                        /* not a broadcast packet */
                        if (bcmp ((caddr_t) eptr->ether_shost,
                                  (caddr_t) &sc->is_addr[0], 6) == 0) {
                                /* we sent this multicast packet -
                                   don't receive it */
                                recvthis = 0;
                        } else {
                                /* Trust higher level code to do the correct
                                   filtering. The ethernet does this too. */
                        }
                }

                if (recvthis) {
                        /* DECnet statistics */
                        if (sc->is_ctrblk.est_mbytercvd + bytes >
                            sc->is_ctrblk.est_mbytercvd) {
                                sc->is_ctrblk.est_mbytercvd += bytes;
                        }
                }
        } else {
                /* Not a mutlicast packet */
                if (bcmp ((caddr_t) eptr->ether_dhost,
                         (caddr_t) &sc->is_addr[0], 6) == 0) {
                        /* this is for us */
                } else if ((sc->data.ifflagscopy & IFF_PROMISC) == 0) {
                        recvthis = 0;
                }
        }
        flags = sc->data.vcs[vc].flags;
        if ((flags & OTTO_VC_UPDATED) == 0 &&
            bcmp ((caddr_t)eptr->ether_shost,
                  (caddr_t)sc->data.vcs[vc].addr, 6) == 0 && !otto_nolearn) {
        	sc->data.vcs[vc].flags =
        		(flags | OTTO_VC_UPDATED) & ~OTTO_VC_STALE;
        }

        /* throw the packet away if we don't want to receive it, unless
           we're acting as a bridge */
        if (!recvthis && !sc->data.otto_isbridging) {
                OTTO_STATS_INC_L (&sc->data, rcv_discard, link);
                return (0);
        }

        eh = *eptr;
        eptr = &eh;
        eptr->ether_type = etype;
        len -= hdrsize;

	/* remove header */
	m_adj (m0, hdrsize);

        /* here we assume that ARP packets fit into a single mbuf */
        /* We cleaned only one mbuf, and need contiguous pkt memory */
        if (etype == ETHERTYPE_ARP && len <= m0->m_len &&
            otto_arp_rcv (sc, mtod (m0, caddr_t), len, vc, eptr)) {
                OTTO_STATS_INC_L (&sc->data, rcv_arp, link);
                return (0);
        }

#ifdef notdef
        anet_setentry (sc->controller,
                (struct anet_etheraddr *) eptr->ether_shost,
                ah->srcesa, (anet_flagstype) ah->flags);
#endif

        if (len <= OTTOMLEN && m0->m_len >= len) {
        	/* packet would fit in one small mbuf --- copy it */
	    	struct mbuf *mtmp;
		OTTOMGET (mtmp);
		if (mtmp != (struct mbuf *) 0) {
			bcopy (mtod (m0, char *), mtod (mtmp, char *), len);
			copied = 1;
			m0 = mtmp;
			m0->m_len = len;
		}
        }

#if (NPFWD > 0)
        {
		int fwded = 0;
		if (sc->otto_isbridging) {
			fwded = pfwd_forward (&sc->is_if, m0, len,
				eptr, recvthis);
		}
		if (!recvthis) {
			if (!fwded && copied) {
				OTTOMFREEM (m0);
			}
			return (fwded && !copied);
		}
        }
#endif 

#ifdef __osf__
        m0->m_pkthdr.rcvif = &sc->is_if;
        m0->m_pkthdr.len = len;
#ifdef AC_IPUP
        ether_input (&sc->is_if, eptr, m0); /*len not used*/
#else
        ether_input (&sc->is_if, eptr, m0, 0); /*len not used*/
#endif
#else
        /* Dispatch this packet */
        net_read (&(sc->is_ed), eptr, m0, len, 0, 0);
#endif

        return (!copied);
}
