#include "otto_platform.h"

static struct mbuf *otto_en_fddi_add (struct otto_softc *sc,
	struct mbuf *m, struct sockaddr *dst,
	unsigned char *osrc, unsigned char *odst, int type,
	unsigned vc, int *plen);
static int otto_en_fddi_read (struct otto_softc *sc, struct mbuf *m0,
        unsigned bytes, unsigned len, unsigned vc, int link);

struct otto_en otto_en_fddi = {
	otto_en_fddi_add,
	otto_en_fddi_read
};

static char llchdr[6] = { LLC_SNAP_LSAP,LLC_SNAP_LSAP,LLC_UI,0,0,0 };

#define HDRSIZEFDDI (2 + 2 + 6 + 6 + sizeof (llchdr) + 2)
	/* size of header for FDDI Gigaswitch encapsulation
	   first 2 bytes = bytes of zeroes in front of Ethernet header
	   second 2 bytes = FDDI stuff?
	   first 6 bytes = dest UID
	   second 6 bytes = source UID
	   third 2 bytes = type field */

static unsigned char bit_reverse[256] = {
	0x00, 0x80, 0x40, 0xc0, 0x20, 0xa0, 0x60, 0xe0,
	0x10, 0x90, 0x50, 0xd0, 0x30, 0xb0, 0x70, 0xf0,
	0x08, 0x88, 0x48, 0xc8, 0x28, 0xa8, 0x68, 0xe8,
	0x18, 0x98, 0x58, 0xd8, 0x38, 0xb8, 0x78, 0xf8,
	0x04, 0x84, 0x44, 0xc4, 0x24, 0xa4, 0x64, 0xe4,
	0x14, 0x94, 0x54, 0xd4, 0x34, 0xb4, 0x74, 0xf4,
	0x0c, 0x8c, 0x4c, 0xcc, 0x2c, 0xac, 0x6c, 0xec,
	0x1c, 0x9c, 0x5c, 0xdc, 0x3c, 0xbc, 0x7c, 0xfc,
	0x02, 0x82, 0x42, 0xc2, 0x22, 0xa2, 0x62, 0xe2,
	0x12, 0x92, 0x52, 0xd2, 0x32, 0xb2, 0x72, 0xf2,
	0x0a, 0x8a, 0x4a, 0xca, 0x2a, 0xaa, 0x6a, 0xea,
	0x1a, 0x9a, 0x5a, 0xda, 0x3a, 0xba, 0x7a, 0xfa,
	0x06, 0x86, 0x46, 0xc6, 0x26, 0xa6, 0x66, 0xe6,
	0x16, 0x96, 0x56, 0xd6, 0x36, 0xb6, 0x76, 0xf6,
	0x0e, 0x8e, 0x4e, 0xce, 0x2e, 0xae, 0x6e, 0xee,
	0x1e, 0x9e, 0x5e, 0xde, 0x3e, 0xbe, 0x7e, 0xfe,
	0x01, 0x81, 0x41, 0xc1, 0x21, 0xa1, 0x61, 0xe1,
	0x11, 0x91, 0x51, 0xd1, 0x31, 0xb1, 0x71, 0xf1,
	0x09, 0x89, 0x49, 0xc9, 0x29, 0xa9, 0x69, 0xe9,
	0x19, 0x99, 0x59, 0xd9, 0x39, 0xb9, 0x79, 0xf9,
	0x05, 0x85, 0x45, 0xc5, 0x25, 0xa5, 0x65, 0xe5,
	0x15, 0x95, 0x55, 0xd5, 0x35, 0xb5, 0x75, 0xf5,
	0x0d, 0x8d, 0x4d, 0xcd, 0x2d, 0xad, 0x6d, 0xed,
	0x1d, 0x9d, 0x5d, 0xdd, 0x3d, 0xbd, 0x7d, 0xfd,
	0x03, 0x83, 0x43, 0xc3, 0x23, 0xa3, 0x63, 0xe3,
	0x13, 0x93, 0x53, 0xd3, 0x33, 0xb3, 0x73, 0xf3,
	0x0b, 0x8b, 0x4b, 0xcb, 0x2b, 0xab, 0x6b, 0xeb,
	0x1b, 0x9b, 0x5b, 0xdb, 0x3b, 0xbb, 0x7b, 0xfb,
	0x07, 0x87, 0x47, 0xc7, 0x27, 0xa7, 0x67, 0xe7,
	0x17, 0x97, 0x57, 0xd7, 0x37, 0xb7, 0x77, 0xf7,
	0x0f, 0x8f, 0x4f, 0xcf, 0x2f, 0xaf, 0x6f, 0xef,
	0x1f, 0x9f, 0x5f, 0xdf, 0x3f, 0xbf, 0x7f, 0xff
};

static struct mbuf *otto_en_fddi_add (struct otto_softc *sc,
	struct mbuf *m, struct sockaddr *dst,
	unsigned char *osrc, unsigned char *odst, int type,
	unsigned vc, int *plen)
{
	unsigned char *padptr;
	unsigned char *da;
	unsigned char *sa;
	unsigned char *lh;
	int len;

	len = m_length (m);
#ifdef __osf__
	M_PREPEND (m, HDRSIZEFDDI, M_DONTWAIT);
	if (m == 0) {
		return ((struct mbuf *) 0);
	}
#else
	if (m->m_off > MMAXOFF ||
	    MMINOFF + HDRSIZEFDDI > m->m_off ||
	    (m->m_off&0x3)!=0) {
	    	struct mbuf *mtmp;
		OTTOMGET (mtmp);
		if (mtmp == 0) {
			OTTOMFREEM (m);
			return ((struct mbuf *) 0);
		}
		mtmp->m_next = m;
		mtmp->m_off = MMAXOFF;
		mtmp->m_len = 0;
		m = mtmp;
	}
	m->m_off -= HDRSIZEFDDI;
	m->m_len += HDRSIZEFDDI;
#endif
	padptr = mtod (m, unsigned char *);
	da = padptr + 4;
	sa = da + 6;
	lh = sa + 6;
	type = htons ((u_short)type);

        /* build the header */
	padptr[0] = 0;
	padptr[1] = 0;
	padptr[2] = 0;
	padptr[3] = 0x54;
        da[0] = bit_reverse[odst[0]];
        da[1] = bit_reverse[odst[1]];
        da[2] = bit_reverse[odst[2]];
        da[3] = bit_reverse[odst[3]];
        da[4] = bit_reverse[odst[4]];
        da[5] = bit_reverse[odst[5]];

        sa[0] = bit_reverse[osrc[0]];
        sa[1] = bit_reverse[osrc[1]];
        sa[2] = bit_reverse[osrc[2]];
        sa[3] = bit_reverse[osrc[3]];
        sa[4] = bit_reverse[osrc[4]];
        sa[5] = bit_reverse[osrc[5]];
	bcopy (llchdr, lh, sizeof (llchdr));
	bcopy ((char *) &type, &lh[sizeof (llchdr)], 2);

	*plen = len + HDRSIZEFDDI;
	return (m);
}

static int otto_en_fddi_read (struct otto_softc *sc, struct mbuf *m0,
        unsigned bytes, unsigned len, unsigned vc, int link)
{
        u_char *p = mtod (m0, unsigned char *);
        u_char *da = p + 4;
        u_char *sa = da + 6;
        u_char *llc = sa + 6;
        u_char *ptype = llc + 6;
        u_short etype;
        struct ether_header eh;
        struct ether_header *eptr = &eh;
        int recvthis;
        int copied = 0;
        int hdrsize = HDRSIZEFDDI - sizeof (llchdr) - 2;
	unsigned flags;

        if (len < HDRSIZEFDDI) {
        	cprintf ("otto_en_fddi_read: rcv pkt too short len=%d bytes=%d\n",
			len, bytes);
        	OTTO_STATS_INC_L (&sc->data, rcv_pkt_too_short, link);
        	return (0);
        }

        /* check the first four bytes.  They should be 0,0,0,0x54 if this is a
           Gigaswitch FDDI packet */
        if (p[0] != 0 || p[1] != 0 || p[2] != 0 || p[3] != 0x54) {
                OTTO_STATS_INC_L (&sc->data, rcv_pkt_not_ethernet, link);
                return (0);
        }

        /* check for llc header */
	if (bcmp (llchdr, llc, sizeof (llchdr)) == 0) {
		/* pick up ethertype */
		etype = (ptype[0] << 8) + ptype[1];
		hdrsize = HDRSIZEFDDI;
	} else {
		etype = 0;
        }

        /* pick up addresses  and type field into fake ethernet header */
        eptr->ether_dhost[0] = bit_reverse[da[0]];
        eptr->ether_dhost[1] = bit_reverse[da[1]];
        eptr->ether_dhost[2] = bit_reverse[da[2]];
        eptr->ether_dhost[3] = bit_reverse[da[3]];
        eptr->ether_dhost[4] = bit_reverse[da[4]];
        eptr->ether_dhost[5] = bit_reverse[da[5]];
        eptr->ether_shost[0] = bit_reverse[sa[0]];
        eptr->ether_shost[1] = bit_reverse[sa[1]];
        eptr->ether_shost[2] = bit_reverse[sa[2]];
        eptr->ether_shost[3] = bit_reverse[sa[3]];
        eptr->ether_shost[4] = bit_reverse[sa[4]];
        eptr->ether_shost[5] = bit_reverse[sa[5]];
	eptr->ether_type = etype;

        /* recvthis is true if we should receive this packet */
        recvthis = 1;
        if ((eptr->ether_dhost[0] & 1) != 0) {
                /* Multicast filter.
                   This is quoted from the RFC on multicast (1112) section
                   7.3 page 9:
                   >> The local network module must not deliver up any 
                   >> multicast packets that were transmitted from that
                   >> module; loopback of multicasts is handled at the IP
                   >> layer or higher.
                   This occurs even if IFF_PROMISC|IFF_ALLMULTI
                   This is the opposite requirement from broadcast loopbacks.
                 */
                if (bcmp ((caddr_t) eptr->ether_dhost,
                          (caddr_t) etherbroadcastaddr, 6) != 0) {
                        /* not a broadcast packet */
                        if (bcmp ((caddr_t) eptr->ether_shost,
                                  (caddr_t) &sc->is_addr[0], 6) == 0) {
                                /* we sent this multicast packet -
                                   don't receive it */
                                recvthis = 0;
                        } else {
                                /* Trust higher level code to do the correct
                                   filtering. The ethernet does this too. */
                        }
                }

                if (recvthis) {
                        /* DECnet statistics */
                        if (sc->is_ctrblk.est_mbytercvd + bytes >
                            sc->is_ctrblk.est_mbytercvd) {
                                sc->is_ctrblk.est_mbytercvd += bytes;
                        }
                }
        } else {
                /* Not a mutlicast packet */
                if (bcmp ((caddr_t) eptr->ether_dhost,
                         (caddr_t) &sc->is_addr[0], 6) == 0) {
                        /* this is for us */
                } else if ((sc->data.ifflagscopy & IFF_PROMISC) == 0) {
                        recvthis = 0;
                }
        }
        flags = sc->data.vcs[vc].flags;
        if ((flags & OTTO_VC_UPDATED) == 0 &&
            bcmp ((caddr_t)eptr->ether_shost,
                  (caddr_t)sc->data.vcs[vc].addr, 6) == 0) {
                sc->data.vcs[vc].flags =
                	(flags | OTTO_VC_UPDATED) & ~OTTO_VC_STALE;
        }

        /* throw the packet away if we don't want to receive it, unless
           we're acting as a bridge */
        if (!recvthis && !sc->data.otto_isbridging) {
                OTTO_STATS_INC_L (&sc->data, rcv_discard, link);
                return (0);
        }

        len -= hdrsize;

	/* remove header */
	m_adj (m0, hdrsize);


        /* here we assume that ARP packets fit into a single mbuf */
        /* We cleaned only one mbuf, and need contiguous pkt memory */
        if (etype == ETHERTYPE_ARP && len <= m0->m_len &&
            otto_arp_rcv (sc, mtod (m0, caddr_t), len, vc, eptr)) {
                OTTO_STATS_INC_L (&sc->data, rcv_arp, link);
                return (0);
        }

#ifdef notdef
        anet_setentry (sc->controller,
                (struct anet_etheraddr *) eptr->ether_shost,
                ah->srcesa, (anet_flagstype) ah->flags);
#endif

        if (len <= OTTOMLEN && m0->m_len >= len) {
        	/* packet would fit in one small mbuf --- copy it */
	    	struct mbuf *mtmp;
		OTTOMGET (mtmp);
		if (mtmp != (struct mbuf *) 0) {
			bcopy (mtod (m0, char *), mtod (mtmp, char *), len);
			copied = 1;
			m0 = mtmp;
			m0->m_len = len;
		}
        }

#if (NPFWD > 0)
        {
		int fwded = 0;
		if (sc->otto_isbridging) {
			fwded = pfwd_forward (&sc->is_if, m0, len,
				eptr, recvthis);
		}
		if (!recvthis) {
			if (!fwded && copied) {
				OTTOMFREEM (m0);
			}
			return (fwded && !copied);
		}
        }
#endif 

#ifdef __osf__
        m0->m_pkthdr.rcvif = &sc->is_if;
        m0->m_pkthdr.len = len;
#ifdef AC_IPUP
        ether_input (&sc->is_if, eptr, m0); /*len not used*/
#else
        ether_input (&sc->is_if, eptr, m0, 0); /*len not used*/
#endif
#else
        /* Dispatch this packet */
        net_read (&(sc->is_ed), eptr, m0, len, 0, 0);
#endif

        return (!copied);
}
