#include "otto_platform.h"

static struct mbuf *otto_en_llc_add (struct otto_softc *sc,
	struct mbuf *m, struct sockaddr *dst,
	unsigned char *osrc, unsigned char *odst, int type,
	unsigned vc, int *plen);
static int otto_en_llc_read (struct otto_softc *sc, struct mbuf *m0,
        unsigned bytes, unsigned len, unsigned vc, int link);

struct otto_en otto_en_llc = {
	otto_en_llc_add,
	otto_en_llc_read
};


static char llchdr[6] = { LLC_SNAP_LSAP,LLC_SNAP_LSAP,LLC_UI,0,0,0 };

#define HDRSIZELLC (sizeof (llchdr) + 2)
	/* size of header for atm forum llc encapsulation
	   2 bytes = type field */

int otto_karp48 = 0;

static struct mbuf *otto_en_llc_add (struct otto_softc *sc,
	struct mbuf *m, struct sockaddr *dst,
	unsigned char *osrc, unsigned char *odst, int type,
	unsigned vc, int *plen)
{
	char *lh;
	int len;

	len = m_length (m);
#ifdef __osf__
	M_PREPEND (m, HDRSIZELLC, M_DONTWAIT);
	if (m == 0) {
		return ((struct mbuf *) 0);
	}
#else
	if (m->m_off > MMAXOFF ||
	    MMINOFF + HDRSIZELLC > m->m_off ||
	    (m->m_off&0x3)!=0) {
	    	struct mbuf *mtmp;
		OTTOMGET (mtmp);
		if (mtmp == 0) {
			OTTOMFREEM (m);
			return ((struct mbuf *) 0);
		}
		mtmp->m_next = m;
		mtmp->m_off = MMAXOFF;
		mtmp->m_len = 0;
		m = mtmp;
	}
	m->m_off -= HDRSIZELLC;
	m->m_len += HDRSIZELLC;
#endif
	lh = mtod (m, char *);
	type = htons ((u_short)type);
        /* build the header */
	bcopy (llchdr, lh, sizeof (llchdr));
	bcopy ((char *) &type, &lh[sizeof (llchdr)], 2);

#ifdef KARP48
	if (otto_karp48) {
	  struct mbuf *mtmp, *mnew;
	  int padval = 0xdeadbeef;
	  char *dstptr;

	  mtmp = m;
	  while (mtmp->m_next)
	    mtmp = mtmp->m_next;
	  
	  if (((len + HDRSIZELLC) % 48) == 0) {
	    if (M_TRAILINGSPACE(mtmp) < sizeof(padval)) {
	      OTTOMGET(mnew);
	      mtmp->m_next = mnew;
	      mnew->m_len = 0;
	      mtmp = mnew;
	    }
	    dstptr = mtod(mtmp, char *) + mtmp->m_len;
	    bcopy((char *) &padval, dstptr, sizeof(padval));
	    mtmp->m_len += 4;
	    m->m_pkthdr.len += 4;
	  }
	  len += 4;
	}
#endif
	*plen = len + HDRSIZELLC;
	return (m);
}


static int otto_en_llc_read (struct otto_softc *sc, struct mbuf *m0,
        unsigned bytes, unsigned len, unsigned vc, int link)
{
        u_char *lh = mtod (m0, unsigned char *);
        struct ether_header eh;
        struct ether_header *eptr = &eh;
        u_short etype = 0;
        int copied = 0;
        int hdrsize = 0;

        if (len < HDRSIZELLC) {
        	cprintf ("otto_en_llc_read: rcv pkt too short len=%d bytes=%d\n",
			len, bytes);
        	OTTO_STATS_INC_L (&sc->data, rcv_pkt_too_short, link);
        	return (0);
        }

	if (bcmp (llchdr, lh, sizeof (llchdr)) == 0) {
		/* pick up ethertype */
		lh += sizeof (llchdr);
		etype = (lh[0] << 8) + lh[1];
		hdrsize = sizeof (llchdr) + 2;
	}

        bcopy ((char *) &sc->is_addr[0], (char *) eptr->ether_dhost, 6);
        bzero ((char *) eptr->ether_shost, 6);
        eptr->ether_type = etype;
        len -= hdrsize;

	/* remove header */
	m_adj (m0, hdrsize);

        /* here we assume that ARP packets fit into a single mbuf */
        /* We cleaned only one mbuf, and need contiguous pkt memory */
        if (etype == ETHERTYPE_ARP && len <= m0->m_len &&
            otto_arp_rcv (sc, mtod (m0, caddr_t), len, vc, eptr)) {
                OTTO_STATS_INC_L (&sc->data, rcv_arp, link);
                return (0);
        }

#ifdef notdef
        anet_setentry (sc->controller,
                (struct anet_etheraddr *) eptr->ether_shost,
                ah->srcesa, (anet_flagstype) ah->flags);
#endif

        if (len <= OTTOMLEN && m0->m_len >= len) {
        	/* packet would fit in one small mbuf --- copy it */
	    	struct mbuf *mtmp;
		OTTOMGET (mtmp);
		if (mtmp != (struct mbuf *) 0) {
			bcopy (mtod (m0, char *), mtod (mtmp, char *), len);
			copied = 1;
			m0 = mtmp;
			m0->m_len = len;
		}
        }

#ifdef __osf__
        m0->m_pkthdr.rcvif = &sc->is_if;
        m0->m_pkthdr.len = len;
#ifdef AC_IPUP
        ether_input (&sc->is_if, eptr, m0); /*len not used*/
#else
        ether_input (&sc->is_if, eptr, m0, 0); /*len not used*/
#endif
#else
        /* Dispatch this packet */
        net_read (&(sc->is_ed), eptr, m0, len, 0, 0);
#endif

        return (!copied);
}
