/* Fake SNMP code. 
   Sends small, sensible, easy to parse packets to transfer information
   that would be sent via huge, stupid, inpenetrable SNMP packets
   if we had a real SNMP implementation. */

#include "otto_platform.h"

int otto_fs_dofm = 1;           /* do FLOWmaster - set to 0 only for testing */
int otto_fs_dorvc = 1;          /* do RVCs - set to 0 only for testing */

/* send a fake SNMP packet. pkttype should be OTTO_FS_REQUEST or OTTO_FS_REPLY
   Called with softc lock held. */
static void sendfakesnmp (struct otto_softc *sc, unsigned pkttype)
{
	struct otto_fakesnmp pkt;
	struct sockaddr dst;
	unsigned one = htonl (1);

	bzero ((caddr_t) &pkt, sizeof (pkt));
	pkt.escape = OTTO_FS_ESCAPE;
	pkt.version = OTTO_FS_VERSION;
	pkt.pkttype = htons (pkttype);
	pkt.type = htonl (OTTO_FS_HOST);
	pkt.subtype = htonl (0);
	pkt.maxvcibits = htonl (11);
	bcopy ((caddr_t)&sc->is_addr[0], (caddr_t)pkt.uid, 6);
	pkt.pad[0] = 0;
	pkt.pad[1] = 0;
	if (otto_fs_dofm) {
		pkt.fmsupport = one;
		pkt.resyncVCI = htonl (2);
		pkt.abrbuffers = htonl (sc->data.rxcells - sc->data.noackcells);
		pkt.abrbufpervc = htonl (50);
	}
	if (otto_fs_dorvc) {
		pkt.rvcsupport = one;
		pkt.bcastVCIout = one;
		pkt.bcastVCIin = one;
		pkt.arpVCIout = one;
		pkt.arpVCIin = one;
		pkt.homeVCI = one;
	}

	dst.sa_family = AF_UNSPEC;
	/* addresses don't matter - this is null encapsulation.
	   Type must be 0. */
	bzero ((char *) dst.sa_data, 14);
	otto_simplesendpkt (sc, OTTO_FS_VCI, (unsigned char *) &pkt, 
		sizeof (pkt), &dst);
}

/* If "oldvc == 0", this call does nothing.
   Otherwise, the VC is torn down.
   Called with softc lock held. */
static void dovcteardown (struct otto_softc *sc, unsigned oldvc)
{
	struct otto_vcdata *vcdata = &sc->data.vcs[oldvc];

	if (oldvc != 0) {
		vcdata->flags &= ~(OTTO_VC_ALLOCATED|OTTO_VC_STOPPED|OTTO_VC_FC);
		vcdata->encap = 0;
		vcdata->etype = 0;
		otto_deallocq (&sc->data, vcdata->q);
		vcdata->q = OTTO_IFQ_NORMAL;
		otto_settxvc (&sc->data, oldvc, 0, 0, 0, 0);
	}
}

/* If "newvc == 0", this call does nothing.
   Otherwise, the VC is set up with flow control enabled (fc non-zero)
   of disabled (fc zero), and as and RVC (rcs non-zero) or as a fixed VC
   (rvc zero).
   Called with softc lock held. */
static void dovcsetup (struct otto_softc *sc, unsigned newvc, unsigned fc,
	unsigned rvc)
{
	struct otto_vcdata *vcdata = &sc->data.vcs[newvc];

	if (newvc == 0) {
		return;
	}
	if (vcdata->bandwidth != 0) {
		(void) otto_update_bandwidth (&sc->data, newvc, 0);
	}
	if (vcdata->flags & OTTO_VC_SCHED) {
		otto_settxvc (&sc->data, newvc, 0, -1, -1, -1);
	}
	if (fc) {
		if ((vcdata->flags & OTTO_VC_FC) == 0) {
			otto_settxvc (&sc->data, newvc, 0, 1, 0, 0);
			otto_setrxvc (&sc->data, newvc, 0, 1, -1, 
			    sc->data.rings[sc->data.defaultring].desc);
			vcdata->flags |= OTTO_VC_FC;
		}
	} else {
		if ((vcdata->flags & OTTO_VC_FC) != 0) {
			otto_settxvc (&sc->data, newvc, 0, 0, 0, 0);
			otto_setrxvc (&sc->data, newvc, 0, 1, -1, 
			    sc->data.rings[sc->data.defaultring].desc);
			vcdata->flags &= ~(OTTO_VC_FC|OTTO_VC_STOPPED);
		}
	}
	vcdata->flags &= ~(OTTO_VC_ALLOCATED|OTTO_VC_SCHED|OTTO_VC_NOFC);
	if (rvc) {
		vcdata->encap = 0;
		vcdata->etype = 0;
		otto_deallocq (&sc->data, vcdata->q);
		vcdata->q = OTTO_IFQ_NORMAL;
		vcdata->flags |= OTTO_VC_RVC;
	} else {
		vcdata->encap = OTTO_EN_RAW;
		vcdata->etype = 0;
		otto_deallocq (&sc->data, vcdata->q);
		vcdata->q = OTTO_IFQ_SPECIAL;
		vcdata->flags |= OTTO_VC_FIXED;
	}
}

int otto_nofakesnmp = 0;

/* The link has just come up (up=1) or gone down (up=0).
   Called with softc lock held.  */
void otto_fakesnmp_updown (struct otto_softc *sc, unsigned up)
{
	unsigned i;

	sc->data.snmp.up = (up?1:0);
	if (up && !otto_nofakesnmp) {
		sendfakesnmp (sc, OTTO_FS_REQUEST);
		sc->data.snmp.lastpktsent = time.tv_sec;
	} else {
		sc->data.snmp.lastpktsent = 0;
		sc->data.snmp.lastpktrcvd = 0;
		sc->data.snmp.backoff = 1;
		dovcteardown (sc, sc->data.snmp.pkt.resyncVCI);
		dovcteardown (sc, sc->data.snmp.pkt.bcastVCIout);
		dovcteardown (sc, sc->data.snmp.pkt.bcastVCIin);
		dovcteardown (sc, sc->data.snmp.pkt.arpVCIout);
		dovcteardown (sc, sc->data.snmp.pkt.arpVCIin);
		dovcteardown (sc, sc->data.snmp.pkt.homeVCI);
		if (sc->data.snmp.pkt.fmsupport) {
			for (i = 1; i != OTTO_MAXVC; i++) {
				if (sc->data.vcs[i].flags & OTTO_VC_FC) {
					sc->data.vcs[i].flags &= ~OTTO_VC_FC;
					otto_settxvc (&sc->data, i, -1, 0, -1, -1);
				}
				otto_setrxvc (&sc->data, i, -1, -1, 0, -1);
			}
			fc_restart (sc->data.fcs);
		}
		if (sc->data.snmp.pkt.rvcsupport) {
			for (i = 1; i != OTTO_MAXVC; i++) {
				if (sc->data.vcs[i].flags & OTTO_VC_RVC) {
					struct otto_map_newmapping m;
					sc->data.vcs[i].flags &= ~OTTO_VC_RVC;
					while (otto_map_from_vc (
					    &sc->data.vcmap, i, m.addr)) {
						otto_map_delete (&sc->data.vcmap,
							m.addr);
					}
				}
			}
		}
		bzero ((caddr_t) &sc->data.snmp.pkt, sizeof (sc->data.snmp.pkt));
	}
}

static int insanevc (unsigned vc)
{
	return (vc == 0 || vc >= OTTO_MAXVC);
}

/* A packet has been received on VCI 16.
   The bytes of the packet are p[0...len-1].
   Called with softc lock held.  */
void otto_fakesnmp_rcv (struct otto_softc *sc, unsigned char *p, unsigned len)
{
	struct otto_fakesnmp tmp;
	unsigned *ui;
	unsigned i;
	unsigned oldfmsupport;
	unsigned oldhomeVCI;
	if (otto_nofakesnmp || len < sizeof (struct otto_fakesnmp)) {
		return;
	}
	bcopy ((caddr_t) p, (caddr_t) &tmp, sizeof (tmp));
	if (tmp.escape != OTTO_FS_ESCAPE || tmp.version != OTTO_FS_VERSION) {
		return;
	}
	tmp.pkttype = ntohs (tmp.pkttype);
	if (tmp.pkttype != OTTO_FS_REQUEST && tmp.pkttype != OTTO_FS_REPLY) {
		return;
	}	 
	sc->data.snmp.lastpktrcvd = time.tv_sec;
	sc->data.snmp.backoff = 32;

	tmp.type = ntohl (tmp.type);
	tmp.subtype = ntohl (tmp.subtype);
	tmp.maxvcibits = ntohl (tmp.maxvcibits);
	for (ui = &tmp.fmsupport; ui <= &tmp.maxpvcVCI; ui++) {
		*ui = ntohl (*ui);
	}
	if (insanevc (tmp.resyncVCI)) {
		tmp.fmsupport = 0;
	}
	if (!tmp.fmsupport) {
		tmp.resyncVCI = 0;
		tmp.abrbuffers = 0;
		tmp.abrbufpervc = 0;
		if (sc->data.snmp.pkt.fmsupport) {
			for (i = 1; i != OTTO_MAXVC; i++) {
				if (sc->data.vcs[i].flags & OTTO_VC_FC) {
					sc->data.vcs[i].flags &= ~OTTO_VC_FC;
					otto_settxvc (&sc->data, i, -1, 0, -1, -1);
				}
				otto_setrxvc (&sc->data, i, -1, -1, 0, -1);
			}
			fc_restart (sc->data.fcs);
		}
	}
	if (insanevc (tmp.bcastVCIout) || insanevc (tmp.bcastVCIin) ||
	    insanevc (tmp.arpVCIout) || insanevc (tmp.arpVCIin) ||
	    insanevc (tmp.homeVCI)) {
		tmp.rvcsupport = 0;
    	}
	if (!tmp.rvcsupport) {
		tmp.bcastVCIout = 0;
		tmp.bcastVCIin = 0;
		tmp.arpVCIout = 0;
		tmp.arpVCIin = 0;
		tmp.homeVCI = 0;
		if (sc->data.snmp.pkt.rvcsupport) {
			for (i = 1; i != OTTO_MAXVC; i++) {
				if (sc->data.vcs[i].flags & OTTO_VC_RVC) {
					struct otto_map_newmapping m;
					sc->data.vcs[i].flags &= ~OTTO_VC_RVC;
					while (otto_map_from_vc (
					    &sc->data.vcmap, i, m.addr)) {
						otto_map_delete (&sc->data.vcmap,
							m.addr);
					}
				}
			}
		}
	}
	if (sc->data.snmp.pkt.resyncVCI != tmp.resyncVCI ||
	    sc->data.snmp.pkt.bcastVCIout != tmp.bcastVCIout ||
	    sc->data.snmp.pkt.bcastVCIin != tmp.bcastVCIin ||
	    sc->data.snmp.pkt.arpVCIout != tmp.arpVCIout ||
	    sc->data.snmp.pkt.arpVCIin != tmp.arpVCIin ||
	    sc->data.snmp.pkt.homeVCI != tmp.homeVCI) {
		dovcteardown (sc, sc->data.snmp.pkt.resyncVCI);
		dovcteardown (sc, sc->data.snmp.pkt.bcastVCIout);
		dovcteardown (sc, sc->data.snmp.pkt.bcastVCIin);
		dovcteardown (sc, sc->data.snmp.pkt.arpVCIout);
		dovcteardown (sc, sc->data.snmp.pkt.arpVCIin);
		dovcteardown (sc, sc->data.snmp.pkt.homeVCI);
		dovcsetup (sc, tmp.resyncVCI, 0, 0);
		dovcsetup (sc, tmp.bcastVCIout, tmp.fmsupport, 1);
		dovcsetup (sc, tmp.bcastVCIin, tmp.fmsupport, 1);
		dovcsetup (sc, tmp.arpVCIout, tmp.fmsupport, 1);
		dovcsetup (sc, tmp.arpVCIin, tmp.fmsupport, 1);
		dovcsetup (sc, tmp.homeVCI, tmp.fmsupport, 1);

		if (tmp.bcastVCIout != 0) {
			unsigned addr[OTTO_MAP_NADDR];
			bzero ((char *) addr, sizeof (addr));
			bcopy ((caddr_t)etherbroadcastaddr, (caddr_t)addr, 6);
			otto_map_set (&sc->data.vcmap, addr, tmp.bcastVCIout);
		} else if (sc->data.snmp.pkt.bcastVCIout != 0) {
			unsigned addr[OTTO_MAP_NADDR];
			bzero ((char *) addr, sizeof (addr));
			bcopy ((caddr_t)etherbroadcastaddr, (caddr_t)addr, 6);
			otto_map_delete (&sc->data.vcmap, addr);
		}
    	}
    	oldfmsupport = sc->data.snmp.pkt.fmsupport;
	oldhomeVCI = sc->data.snmp.pkt.homeVCI;
	sc->data.snmp.pkt = tmp;

	/* The home VCI has changed.   Broadcast a SARP reply on the
	   ARP out channel to let others know who we now are. */
	if (oldhomeVCI != tmp.homeVCI && tmp.homeVCI != 0 && tmp.arpVCIout !=0){
		otto_arp_sendsarp (sc, tmp.arpVCIout, 0,
			etherbroadcastaddr, etherbroadcastaddr, 1);
	}

	if (tmp.pkttype == OTTO_FS_REQUEST) {
		sendfakesnmp (sc, OTTO_FS_REPLY);
	}

	if (sc->data.snmp.pkt.fmsupport && !oldfmsupport) {
		for (i = 1; i != OTTO_MAXVC; i++) {
			if ((sc->data.vcs[i].flags&(OTTO_VC_SCHED|OTTO_VC_FIXED|OTTO_VC_NOFC))==0 &&
			    sc->data.vcs[i].bandwidth == 0) {
				sc->data.vcs[i].flags |= OTTO_VC_FC;
				otto_settxvc (&sc->data, i, -1, 1, 0, 0);
			}
			otto_setrxvc (&sc->data, i, -1, -1, 1, -1);
		}
		fc_enable (sc->data.fcs);
	}
}

/* This procedure is called once per second.
   Called with softc lock held.  */
void otto_fakesnmp_tick (struct otto_softc *sc)
{
	time_t now;
	if (otto_nofakesnmp || !sc->data.snmp.up) {
		return;
	}
	now = time.tv_sec;
	if (sc->data.snmp.lastpktrcvd != 0 &&
	    now > sc->data.snmp.lastpktrcvd + 100) {
		/* timeout - we're getting no response.
		   Treat the link like it just died and came back up */
		otto_fakesnmp_updown (sc, 0);
		otto_fakesnmp_updown (sc, 1);
	} else if (now > sc->data.snmp.lastpktsent+sc->data.snmp.backoff) {
		sendfakesnmp (sc, OTTO_FS_REQUEST);
		sc->data.snmp.lastpktsent = time.tv_sec;
		if (sc->data.snmp.backoff < 30) {
			sc->data.snmp.backoff *= 4;
		}
	}
}
