#if !defined (_OTTO_MAP_H_)
#define _OTTO_MAP_H_
/* Map between vcs and other address types. */


#define OTTO_MAP_NADDR	6		/* number of unsigneds in an address */

/* A "struct otto_map" maps between addresses and VCs.
   An entry in the map contains: */
struct otto_map_newmapping {
	unsigned addr[OTTO_MAP_NADDR];		/* an address */
	unsigned vc;			/* a VC */
};
struct otto_map_mapping {		/* used for backwards compatibility */
	unsigned addr[2];		/* an address */
	unsigned vc;			/* a VC */
};

#ifdef KERNEL
struct otto_map_entry;

struct otto_map {
	struct otto_map_entry *list;
	struct otto_map_entry *free;
	struct mbuf *m;
	OTTO_SMP_LOCKTYPE lock;
};

void otto_map_init (struct otto_map *map);
/* Initializes an "struct otto_map" to unlocked and empty. */

void otto_map_set (struct otto_map *map, unsigned addr[], unsigned vc);
/* Search "map" for an entry that has address "addr[0..OTTO_MAP_NADDR-1]".
   If none is found, create one.
   In any case, set the entry's address to "addr[0..OTTO_MAP_NADDR-1]", 
   and its vc to "vc". */

void otto_map_delete (struct otto_map *map, unsigned addr[]);
/* Search "map" for an entry that has address "addr[0..OTTO_MAP_NADDR-1]".
   If one is found, delete it. */

int otto_map_from_vc (struct otto_map *map, unsigned vc, unsigned addr[]);
/* Search "map" for an entry that has vc "vc".
   If such an entry is found, set "addr[0..OTTO_MAP_NADDR-1]" to the
   address in the entry, and return 1.  Otherwise return 0. */

unsigned otto_map_to_vc (struct otto_map *map, unsigned addr[]);
/* Search "map" for an entry that has address "addr[0..OTTO_MAP_NADDR-1]".
   If such an entry is found, return the vc in the entry.  Otherwise return 0.*/

unsigned otto_map_ip_to_vc (struct otto_map *map, unsigned addr[]);
/* Search "map" for an entry such that
		"entry->addr[0] == (entry->addr[1] & addr[0]) &&
		 entry->addr[2] == (entry->addr[3] & addr[1]) &&
		 ...
		 entry->addr[OTTO_MAP_NADDR-2] ==
		    (entry->addr[OTTO_MAP_NADDR-1] & addr[OTTO_MAP_NADDR/2-1])"
   If such an entry is found, return the vc in the entry.  Otherwise return 0.

   This procedure is intended to be used for mapping entire IP networks to VCs.
   */

int otto_map_copy (struct otto_map *map,
	struct otto_map_newmapping mapping[], unsigned count,
	int touser);
/* If there are N entries in the map "map", 
   Copies MIN(N, count) entries from "map"  into "mapping[0..MIN(count-1,N-1)]"
   and returns MAX(N, count) of -1 on error.
   If touser is true, the copy is done with copyout(). */


typedef void *(*otto_iterfunc) (void *arg, struct otto_map_newmapping *mapping);
/* Type of function passed to otto_map_iterate; see below. */

void *otto_map_iterate (struct otto_map *map, otto_iterfunc func, void *arg);
/* For each entry in the map "map", "func" is called with arguments
   "arg" and "mapping", where "mapping" points to the mapping represented
   by the entry.    On any particular entry, if "func" returns 0, the 
   iteration proceeds to the next entry.   If it returns non-zero, 
   the iteration terminates.   If there is no next entry, the iteration
   terminates.  In either case, otto_map_iterate returns the rteturn value
   of the last call to func, or 0 if there were no calls.

   func may not block.
   It is called under the map's internal lock. */
#endif
#endif
