#include "otto_platform.h"

#define OTTO_CLOCK() (ottoclockloc? *ottoclockloc: 0)

void otto_event (struct otto_data *od, unsigned vc,
	unsigned size, unsigned eventtype)
{
	register struct otto_counts *counts = od->counts;

	if (counts != (struct otto_counts *) 0 && eventtype < 2 &&
		vc != 0 && vc < OTTO_MAXVC) {
		register struct otto_perevent_counts *countstotal =
			&counts[0].perevent[eventtype];
		register struct otto_perevent_counts *countsvc =
			&counts[vc].perevent[eventtype];
		register unsigned cells = (size + 47) / 48;

		countsvc->cells += cells;
		countsvc->pkts++;
		countstotal->cells += cells;
		countstotal->pkts++;
	}

	if (od->times.state == OTTO_TIMES_COLLECT &&
	    (eventtype == OTTO_EVENT_TICK ||
	     od->times.vclo <= vc && vc < od->times.vchi)) {
		register unsigned nevents = od->times.nevents;
		register struct otto_event *event = &od->times.events[nevents];
		event->cycles = OTTO_CLOCK ();
		event->len = size;
		event->vctype  = OTTO_EVENT_MAKE (eventtype, vc);
		nevents++;
		od->times.nevents = nevents;
		if (nevents == od->times.maxevents) {
			od->times.state = OTTO_TIMES_HOLD;
			od->times.endcycles = OTTO_CLOCK ();
			microtime (&od->times.endtime);
		}
	}
}

void otto_clear_times (struct otto_times *times)
{
	if (times->state != OTTO_TIMES_NULL) {
		OTTO_KM_FREE (times->events, KM_DEVBUF,
			times->maxevents*sizeof (struct otto_event));
		bzero ((caddr_t) times, sizeof (*times));
		times->state = OTTO_TIMES_NULL;
	}
}


int otto_initialize_times (struct otto_times *times, unsigned maxevents)
{
	if (times->state != OTTO_TIMES_NULL && times->maxevents != maxevents) {
		otto_clear_times (times);
	}
	if (maxevents == 0) {
		return (EINVAL);
	}
	if (maxevents > OTTO_EVENTS_MAX) {
		return (ENOMEM);
	}
	if (times->state == OTTO_TIMES_NULL) {
		KM_ALLOC (times->events, struct otto_event *,
			maxevents*sizeof (struct otto_event),
			KM_DEVBUF, KM_NOARG);
		if (times->events == (struct otto_event *) 0) {
			return (ENOMEM);
		}
		times->maxevents = maxevents;
	}
	times->nevents = 0;
	times->state = OTTO_TIMES_HOLD;
	return (0);
}

int otto_measure_ioctl (struct otto_data *od, struct otto_measure *measure)
{
	unsigned cmd = measure->cmd;
	unsigned vc = measure->vc;

	if (OTTO_CT (cmd) == OTTO_COUNTERS) {
		struct otto_counts *counts = od->counts;

		if (vc >= OTTO_MAXVC) {
			return (EINVAL);
		}
		if (counts == (struct otto_counts *) 0) {
			KM_ALLOC (counts, struct otto_counts *,
				OTTO_MAXVC*sizeof (struct otto_counts),
				KM_DEVBUF, KM_NOARG);
			if (counts == (struct otto_counts *) 0) {
				return (ENOMEM);
			}
			bzero ((char *) counts,
				OTTO_MAXVC*sizeof (struct otto_counts));
			od->counts = counts;
		}
		switch (OTTO_C_ACT (cmd)) {
		case OTTO_C_ZERO:
			measure->uu.counts = counts[vc];
			bzero ((char *)&counts[vc],sizeof (struct otto_counts));
			return (0);
		case OTTO_C_GET:
			measure->uu.counts = counts[vc];
			return (0);
		}
	} else if (OTTO_CT (cmd) == OTTO_TIMERS) {
		unsigned state = od->times.state;
		int error;
		measure->vc = 0;
		switch (OTTO_T_ACT (cmd)) {
		case OTTO_T_START:
			error = otto_initialize_times (&od->times,
					measure->uu.times.maxevents);
			if (error != 0) {
				return (error);
			}
			od->times.vclo = measure->uu.times.vclo;
			od->times.vchi = measure->uu.times.vchi;
			od->times.endcycles = 0;
			od->times.endtime.tv_sec = 0;
			od->times.endtime.tv_usec = 0;
			microtime (&od->times.starttime);
			od->times.startcycles = OTTO_CLOCK ();
			od->times.state = OTTO_TIMES_COLLECT;
			measure->uu.times = od->times;
			measure->uu.times.events = (struct otto_event *)0;
			return (0);
		case OTTO_T_STOP:
			if (state == OTTO_TIMES_COLLECT) {
				od->times.state = OTTO_TIMES_HOLD;
				od->times.endcycles = OTTO_CLOCK ();
				microtime (&od->times.endtime);
				measure->uu.times = od->times;
				measure->uu.times.events=(struct otto_event *)0;
				return (0);
			}
			return (EINVAL);
		case OTTO_T_GET:
			if (state != OTTO_TIMES_NULL) {
				unsigned nevents = measure->uu.times.maxevents;
				struct otto_event *events =
					measure->uu.times.events;
				if (od->times.nevents < nevents) {
					nevents = od->times.nevents;
				}
				error = copyout ((caddr_t) od->times.events,
				   (caddr_t) events,
				   nevents*sizeof (struct otto_event));
				if (error == 0) {
					if (state == OTTO_TIMES_COLLECT) {
						od->times.endcycles =
							OTTO_CLOCK ();
						microtime (&od->times.endtime);
					}
					measure->uu.times = od->times;
					measure->uu.times.events = events;
					if (state == OTTO_TIMES_COLLECT) {
						microtime (&od->times.starttime);
						od->times.startcycles =
							OTTO_CLOCK ();
						od->times.nevents = 0;
					}
				}
				return (error);
			}
			return (EINVAL);
		case OTTO_T_CLEAR:
			otto_clear_times (&od->times);
			return (0);
		}
	}
	printf ("otto_measure_ioctl: unknown command %x\n", cmd);
	return (EINVAL);
}
