#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <strings.h>
#include "if_ottomaint.h"

static void printbothlinks (FILE *fp, struct otto_stats_bothlinks *bothlinks);
static void printlink (FILE *fp, int link, struct otto_stats_perlink *bothlinks);
static void printestat (FILE *fp, struct estat *estat);

static void usage ()
{
	fprintf (stderr, "usage: ottogetcounts [-zs] [interface]\n");
	exit (2);
}

main (argc, argv)
int argc;
char *argv[];
{
	int s;
	struct ctrreq cntrs;
	struct otto_stats_bothlinks *bothlinks;
	struct otto_stats_perlink *link;
	int i;
	char *flags;
	char *interface = "otto0";
	int ioc = SIOCRDCTRS;
	char *sioc = "SIOCRDCTRS";
	int argn;
	char err[100];
	int estat = 0;

	for (argn = 1; argn < argc && argv[argn][0] == '-'; argn++) {
		flags = argv[argn];
		for (i = 1; flags[i] != '\0'; i++) {
			switch (flags[i])
			{
			case 's':
				estat = 1;
				break;
			case 'z':
				ioc = SIOCRDZCTRS;
				sioc = "SIOCRDZCTRS";
				break;
			default:  usage ();
			}
		}
	}

	if (argn < argc) {
		interface = argv[argn++];
	}

	if (argn < argc) {
		usage ();
	}

	(void) sprintf (err, "ottogetcounts: %s: %s", interface, sioc);

	s = socket (AF_INET, SOCK_DGRAM, 0);
	if (s == -1) {
		perror ("ottogetcounts: socket");
		exit (2);
	}

	if (estat) {
		(void) strcpy (cntrs.ctr_name, interface);
		cntrs.ctr_type = CTR_ETHER;
		if (ioctl (s, ioc, (char *) &cntrs) == -1) {
			perror (err);
			exit (2);
		}
		printestat (stdout, &cntrs.ctr_ether);
	}

	(void) strcpy (cntrs.ctr_name, interface);
	cntrs.ctr_type = OTTO_CTR_BOTHLINKS;
	if (ioctl (s, ioc, (char *) &cntrs) == -1) {
		perror (err);
		exit (2);
	}
	bothlinks = (struct otto_stats_bothlinks *) &cntrs.ctr_ctrs;
	printbothlinks (stdout, bothlinks);

	(void) strcpy (cntrs.ctr_name, interface);
	cntrs.ctr_type = OTTO_CTR_LINK0;
	if (ioctl (s, ioc, (char *) &cntrs) == -1) {
		perror (err);
		exit (2);
	}
	link = (struct otto_stats_perlink *) &cntrs.ctr_ctrs;
	printlink (stdout, 0, link);

	(void) strcpy (cntrs.ctr_name, interface);
	cntrs.ctr_type = OTTO_CTR_LINK1;
	if (ioctl (s, ioc, (char *) &cntrs) == -1) {
		perror (err);
		exit (2);
	}
	link = (struct otto_stats_perlink *) &cntrs.ctr_ctrs;
	printlink (stdout, 1, link);

	(void) close (s);

	exit (0);
}

#define printfield(fp, bothlinks, f) do {				\
		char *p = "f";						\
		while (*p == ' ') {					\
			p++;						\
		}							\
		(void) fprintf ((fp), "%-25s %10u\n", p, (bothlinks)->f);\
	} while (0)

static void printbothlinks (FILE *fp, struct otto_stats_bothlinks *bothlinks)
{
	printfield (fp, bothlinks, reset);
	printfield (fp, bothlinks, xmt_added_mbuf);
	printfield (fp, bothlinks, xmt_unicast_too_big);
	printfield (fp, bothlinks, xmt_multicast_too_big);
	printfield (fp, bothlinks, xmt_too_many_xmt_frags);
	printfield (fp, bothlinks, xmt_bad_aligned_frag);
	printfield (fp, bothlinks, xmt_bad_len_frag);
	printfield (fp, bothlinks, xmt_short_frag);
	printfield (fp, bothlinks, simple_send);
	printfield (fp, bothlinks, status_change);
	printfield (fp, bothlinks, no_free_desc);
	printfield (fp, bothlinks, too_many_loops);
	printfield (fp, bothlinks, tran_error);
	printfield (fp, bothlinks, dma_error);
	printfield (fp, bothlinks, schedxmt);
	printfield (fp, bothlinks, zapcache);
	printfield (fp, bothlinks, rcvsarp);
	printfield (fp, bothlinks, sendsarp);
}

#define printlinkfield(fp, l, link, f) do {				\
		char buf[100];						\
		char *p = "f";						\
		while (*p == ' ') {					\
			p++;						\
		}							\
		(void) sprintf (buf, "%s%d", p, (l));			\
		(void) fprintf ((fp), "%-25s %10u\n", buf, (link)->f);	\
	} while (0)

static void printlink (FILE *fp, int l, struct otto_stats_perlink *link)
{
	printlinkfield (fp, l, link, xmt_error);
	printlinkfield (fp, l, link, xmt_dma_error);
	printlinkfield (fp, l, link, xmt_frag_oflow);
	printlinkfield (fp, l, link, rcv_dma_error);
	printlinkfield (fp, l, link, rcv_overflow);
	printlinkfield (fp, l, link, rcv_error);
	printlinkfield (fp, l, link, rcv_crc_error);
	printlinkfield (fp, l, link, rcv_violation);
	printlinkfield (fp, l, link, rcv_frmxcrc);
	printlinkfield (fp, l, link, rcv_frame_err);
	printlinkfield (fp, l, link, rcv_len_err);
	printlinkfield (fp, l, link, rcv_discard);
	printlinkfield (fp, l, link, xmt_ring_full);
	printlinkfield (fp, l, link, saw_syntax_error);
	printlinkfield (fp, l, link, saw_violation);
	printlinkfield (fp, l, link, missed_pkt);
	printlinkfield (fp, l, link, rcv_pkt_too_short);
	printlinkfield (fp, l, link, rcv_pkt_too_large);
	printlinkfield (fp, l, link, rcv_portinfo);
	printlinkfield (fp, l, link, rcv_arp);
	printlinkfield (fp, l, link, rcv_bad_align);
	printlinkfield (fp, l, link, rcv_pkt_not_ethernet);
}

static void printestat (FILE *fp, struct estat *estat)
{
	printfield (fp, estat, est_seconds);
	printfield (fp, estat, est_bytercvd);
	printfield (fp, estat, est_bytesent);
	printfield (fp, estat, est_blokrcvd);
	printfield (fp, estat, est_bloksent);
	printfield (fp, estat, est_mbytercvd);
	printfield (fp, estat, est_mblokrcvd);
	printfield (fp, estat, est_deferred);
	printfield (fp, estat, est_single);
	printfield (fp, estat, est_multiple);
	printf (fp, "%-25s %x\n", "est_sendfail_bm", estat->est_sendfail_bm);
	printfield (fp, estat, est_sendfail);
	printfield (fp, estat, est_collis);
	printfield (fp, estat, est_recvfail);
	printf (fp, "%-25s %x\n", "est_recvfail_bm", estat->est_recvfail_bm);
	printfield (fp, estat, est_unrecog);
	printfield (fp, estat, est_overrun);
	printfield (fp, estat, est_sysbuf);
	printfield (fp, estat, est_userbuf);
	printfield (fp, estat, est_mbytesent);
	printfield (fp, estat, est_mbloksent);
}
